# Narzędzia i Aplikacje Jakarta EE

## Context and Dependency Injection

Michał Wójcik

---

### Wstrzykiwanie zależności

```java
public class EmailUserNotifier {

    public void notifyUser(User user, String msg) {
        //some epice mail sending code
    }

}
```

```java
public class LibraryService {

    @Inject
    private EmailUserNotifier notifier;

}
```

Podstawowe wstrzykiwanie realizowane jest na podstawie typu.

---

### Kilka implementacji

```java
public interface UserNotifier {

    void notifyUser(User user, String msg);

}
```

```java
public class EmailUserNotifier implements UserNotifier {

    public void notifyUser(User user, String msg) {
        //some epice mail sending code
    }

}
```

```java
public class SmsUserNotifier implements UserNotifier {

    public void notifyUser(User user, String msg) {
        //some epice sms sending code
    }

}
```

---

### Problem wybrania implementacji

```java
public class LibraryService {

    @Inject
    private UserNotifier notifier;

}
```

Bez rozróżnienia implementacji otrzymamy
błąd: `DeploymentException: WELD-001409 Injection point has ambiguous dependencies.`

Rozwiązaniem może być niekorzystanie z uogólnionych typów, jednak nie jest to dobre podejście pod względem jakości kodu.

---

### Rozróżnianie po nazwie

Poszczególnym implementacjom można przypisać nazwy:

```java
@Named("emailNotifier")
public class EmailUserNotifier implements UserNotifier {
    //...
}
```

```java
@Named("smsNotifier")
public class SmsUserNotifier implements UserNotifier {
    //...
}
```

Wstrzykując, używamy tej samej adnotacji `@Named`:

```java
public class LibraryService {

    @Inject
    @Named("emailNotifier")
    private UserNotifier notifier;

}
```

Brak kontroli poprawności nazw na etapie kompilacji.

---

### Implementacje domyślne

Poszczególne implementacje można zdefiniować jako domyślne albo alternatywne:

```java
@Alternative
public class EmailUserNotifier implements UserNotifier {
    //...
}
```

```java
@Default
public class SmsUserNotifier implements UserNotifier {
    //...
}
```

```java
public class LibraryService {

    @Inject
    @Default
    private UserNotifier notifier;

}
```

Domyślnie każdy bean CDI jest oznaczony jako `@Default`. Beanów oznaczonych jako `@Alternative` nie można wstrzyknąć
bez konfiguracji w deskryptorze.

---

### Wstrzykiwanie alternatywnych implementacji

Nadpisanie domyślnej implementacji jest możliwe za pomocą deskryptora `beans.xml`:

```xml
<alternatives>
    <class>library.EmailUserNotifier</class>
</alternatives>
```

---

### Rozróżnianie poprzez kwalifikatory

Możemy utworzyć zestaw własnych adnotacji wykorzystywanych do kwalifikowania klasy:

```java
@EmailNotifier
public class EmailUserNotifier implements UserNotifier {
    //...
}
```

```java
@SmsNotifier
public class SmsUserNotifier implements UserNotifier {
    //...
}
```

Wstrzykując, używamy konkretnego samego kwalifikatora:

```java
public class LibraryService {

    @Inject
    @EmailNotifier
    private UserNotifier notifier;

}
```

---

### Definicja kwalifikatora

Definicja kwalifikatora za pomocą adnotacji.

```java
@Qualifier
@Retention(RetentionPolicy.RUNTIME)
@Target({ElementType.FIELD, ElementType.METHOD, ElementType.TYPE})
public @interface EmailNotifier {
}
```

Kontrola na etapie kompilacji, podpowiadanie składni.

---

### Kwalifikator `@New`

Użycie kwalifikatora `@New` przy wstrzykiwaniu, wymusza wstrzyknięcie instancji beana o zasięgu `@Dependent` niezależnie
od tego, jaki zasięg ma przypisana klasa tego beana.

Uwaga, kwalifikator usunięty ze specyfikacji począwszy od CDI 4.0 (Jakarta EE 10). Zasięgi powinny być definiowane po
stronie definiującej (bean lub producent) nie po stronie wstrzyknięcia.

---

### Kwalifikator `@New` - przykład

Przykładowo `UserInfo` jest beanem o zasięgu sesji:

```java
@SessionScoped
public class UserInfo implements Serializable {

}
```

```java
public class BookManager {
    
    @Inject
    private UserInfo info;//the same instance
  
}
```

```java
public class NewspaperManager {
    
    @Inject
    private UserInfo info;//the same instance
  
}
```

```java
public class ClassifiedBookManager {
    
    @Inject
    @New
    private UserInfo info;//new instance
  
}
```

---

### Producenci - motywacja

Motywacje dla użycia **producentów**:

* kontener zależności jest w stanie utworzyć instancje zależności, które:
  * posiadają konstruktor domyślny,
  * lub posiadają konstruktor opatrzony adnotacją `@Inject` i wszystkie ich zależności wstrzykiwane przy pomocy
      konstruktora, są możliwe do utworzenia;
* wiele klas nie spełnia powyższych warunków:
  * szczególnie klasy z bibliotek zewnętrznych.

---

### Producenci

Producenci beanów CDI:

* CDI pozwala na tworzenie własnych fabryk dla wstrzykiwanych beanów,
* przydatne, gdy wstrzykiwany obiekt wymaga złożonej inicjalizacji,
* metoda wytwórcza lub pole klasy muszą być opatrzone adnotacją `@Produces`,
* do metod wytwórczych można stosować kwalifikatory tak jak do klas,
* obiekt zwrócony z metody wytwórczej jest zarządzany przez kontener zależności.

---

### Producenci - przykład

Dodatkowe wymagania klasy `EmailUserNotifier`:

* konieczność podania adresu serwera SMTP i numeru portu, na którym on nasłuchuje,
* konieczność ustawienia loginu i hasła dostępowego,
* konieczność ustawienia domyślnego szablonu.

```java
public class EmailUserNotifier implements UserNotifier {

    public EmailUserNotifier(String smtpAddress, String smtpPort) {
        //
    }

    public void setDefaultTemplate(String template) {
        //
    }

    public void setAuthorization(String login, String password) {
        //
    }

    public void notifyUser(User user, String msg) {
        //some epice mail sending code
    }

}
```

---

### Producenci – przykład

Konstrukcja funkcjonalnego obiektu `EmailUserNotifier`:

```jshelllanguage
EmailUserNotifier notifier = new EmailUserNotifier("192.168.1.17", "991");
notifier.setAuth(user, password);
notifier.setDefaultTemplate("cobalt");
```

Dostarczenie zależności:

* kontener zależności nie jest w stanie samodzielnie przygotować funkcjonalnej instancji klasy `EmailUserNotifier`,
* metoda wytwórcza posłuży kontenerowi jako fabryka obiektów:

```jshelllanguage
@Produces
public EmailUserNotifier createNotifier() {
    EmailUserNotifier notifier = new EmailUserNotifier("192.168.1.17", "991");
    notifier.setAuth(user, password);
    notifier.setDefaultTemplate("cobalt");
    return notifier;
}
```

---

### Producenci – przykład - kwalifikatory

Korzystanie z nazwy:

```jshelllanguage
@Produces
@Named("emailNotifier")
public EmailUserNotifier createNotifier() {
    //...    
}
```

Korzystanie z kwalifikatorów:

```jshelllanguage
@Produces
@EmailNotifier
public EmailUserNotifier createNotifier() {
    //...    
}
```

---

### Producenci - inny przykład

```java
@ApplicationScoped
public class PersistenceUnitManager implements Serializable {

    @PersistenceUnit
    private EntityManagerFactory emFactory;

    @Produces
    @RequestScoped
    public EntityManager createEntityManager() {
        if (emFactory == null) {
            emFactory = Persistence.createEntityManagerFactory("main");
        }
        return emFactory.createEntityManager();
    }

}
```

---

### Producenci - metody zwalniające zasoby

Zwalnianie zasobów:

* podobnie jak utworzenie obiektu może wymagać dodatkowych czynności inicjalizujących, tak zwolnienie zasobów obiektu
  może wymagać dodatkowych czynności sprzątających,
* dodatkowe operacje mogą być przeprowadzone w metodzie, której argument jest opatrzony adnotacją `@Disposes`:

```jshelllanguage
public void closeEntityManager(@Disposes EntityManager em) {
    em.close();
}
```

---

### Rejestrowanie obiektów zarządzanych

Adnotacja `@Prodcues` pozwala także na zarejestrowanie w kontenerze CDI innych mechanizmów Jakarta EE:

* zasoby wstrzykiwane przez `@Resource`,
* fabryka obiektów `EntityManager` wstrzykiwana przez `@PersistenceContext`,
* obiekt `EntityManager` wstrzykiwany przez `@PersistenceUnit`,
* zdalny EJB wstrzykiwany przez `@EJB`,
* usługa sieciowa wstrzykiwana przez `@WebServiceRef`.

---

### Rejestrowanie obiektów zarządzanych cd

```jshelllanguage
@Produces @WebServiceRef(lookup="java:app/service/PaymentService")
PaymentService paymentService;
```

```jshelllanguage
@Produces @EJB(ejbLink="../their.jar#PaymentService")
PaymentService paymentService;
```

```jshelllanguage
@Produces @Resource(lookup="java:global/env/jdbc/CustomerDatasource")
@CustomerDatabase Datasource customerDatabase;
```

```jshelllanguage
@Produces @PersistenceContext(unitName="CustomerDatabase")
@CustomerDatabase EntityManager customerDatabasePersistenceContext;
```

```jshelllanguage
@Produces @PersistenceUnit(unitName="CustomerDatabase")
@CustomerDatabase EntityManagerFactory customerDatabasePersistenceUnit;
```

---

### Callbacki @PostConstruct i @PreDestroy

Dodatkowe czynności inicjujące i zwalniające zasoby mogą być umieszczone w ciele klasy, jeśli dostępny jest jej kod
źródłowy. Służą do tego adnotacje związanie z cyklem życia obiektów:

* `@PostConstruct` - do oznaczenia metody, która zostanie wykonana po utworzeniu beana i wstrzyknięciu wszystkich jego
  zależności,
* `@PreDestroy` - do oznaczenia metody, która zostanie wykonana przed zniszczeniem instancji beana.

```java
public class LibraryService {

    @Inject
    @EmailNotifier
    private UserNotifier notifier;

    @PostConstruct
    public void init() {
        //
    }

    @PreDestroy
    public void destroy() {
        //
    }

}
```

---

### Gdzie wstrzykiwać?

Zależności mogą być wstrzykiwane do:

* innych zależności – oczywiste,
* wszystkich obiektów zarządzanych przez kontener zależności – też oczywiste,
* servletów i filtrów,
* wyrażeń Expression Language (Weld resolver dla EL),
* beanów zarządzanych (`@ManagedBean`) JSF,
* beanów EJB.

---

### Problem konwerterów JSF

Wykorzystanie standardowych konwerterów JSF:

* w JSF 1.x nie można wstrzykiwać zasobów do konwerterów,
* w JSF 2.0 problem ten nie został rozwiązany,
* w JSF 2.3 problem został rozwiązany poprzez `@FacesConverter(managed = true)`.
* instancja konwertera jest tworzona gdzieś w cyklu życia JSF.

Ominięcie problemu poprzez realizację konwerterów jako beanów `@Named` i podawanie ich jawnie w kodzie JSF.

---

### Ręczne odpytywanie kontenera CDI

W miejscach, gdzie nie można wstrzykiwać zasobów w sposób deklaratywny (adnotacja `@Inject`) można programistycznie
odpytywać kontener CDI:

* obiekt `BeanManager` - pozwala na odpytywanie kontenera,
* dostępny przez:
  * JNDI,
  * `ServletContext`, `FacesContext`.

---

### Pobieranie obiektu BeanManager

Pobranie poprzez JDNI (CDI 1.0+):

```jshelllanguage
public BeanManager getBeanManager() {
    try {
        InitialContext initialContext = new InitialContext();
        BeanManager bm = (BeanManager) initialContext.lookup("java:comp/BeanManager");
        return bm;
    } catch (NamingException ex){
        return null;
    }
}
```

Pobranie z `FacesContext` (CDI 1.0+):

```jshelllanguage
public BeanManager getBeanManager() {
    ServletContext context = facesContext.getExternalContext().getContext();
    Object obj = context.getAttribute("jakarta.enterprise.inject.spi.BeanManager");
    BeanManager bm = (BeanManager) obj;
    return bm;
}
```

Wygodnie (CDI 1.1+):

```jshelllanguage
public BeanManager getBeanManager() {
    return CDI.current().getBeanManager();
}
```

---

### Ręczne odpytywanie

Pobranie na podstawie typu i kwalifikatorów:

```jshelllanguage
public<T> T getCDIBean(Class<T> beanClass, Annotation...qualifiers) {
    BeanManager bm = getBeanManager();
    Bean<T> bean = (Bean<T>) bm.getBeans(beanClass, qualifiers).iterator().next();
    CreationalContext<T> ctx = bm.createCreationalContext(bean);
    T beanInstance = (T) bm.getReference(bean, beanClass, ctx);
    return beanInstance;
}
```

Pobranie beanów oznaczonych adnotacją `@Named`:

```jshelllanguage
public Object getCDIBeanByName(String name) {
    BeanManager bm = getBeanManager();
    Bean bean = bm.getBeans(name).iterator().next();
    CreationalContext ctx = bm.createCreationalContext(bean);
    Object o = bm.getReference(bean,bean.getClass(),ctx);
    return o;
}
```

---

### CDI – więcej niż wstrzykiwanie zależności

CDI poza wstrzykiwaniem zależności definiuje również:

* model zdarzeń,
* przecięcia (interceptors),
* dekoratory,
* stereotypy.

Wszystkie te mechanizmy mają na celu rozluźnienie powiązań między komponentami.

---

### Model zdarzeń

Model zdarzeń CDI:

* implementacja wzorca Obserwator,
* model stworzony z myślą o rozluźnieniu powiązań pomiędzy producentami zdarzeń i ich konsumentami,
* eliminuje "bagaż składniowy" charakterystyczny dla detektorów zdarzeń w Java SE.

Obsługa zdarzeń w Swingu:

```jshelllanguage
jButton.addActionListener(new java.awt.event.ActionListener() {

    @Override 
    public void actionPerformed(java.awt.event.ActionEvent evt) {
        //...
    }

});
```

---

### Obiekt Event i obiekt opisu zdarzenia

Opis zdarzenia:

* obiekt generycznej klasy `Event` reprezentuje zdarzenie, które może zostać uruchomione (`fire`),
* opis zdarzenia reprezentuje osobna klasa - parametr typowy dla obiektu zdarzenia (np. `Event<BookReservation>`),
* obiekt klasy `Event` wstrzykujemy do klasy, która będzie wywoływać zdarzenie,
* metoda nasłuchująca rejestruje się jako obserwator w kontenerze CDI, a nie w klasie generującej zdarzenia,
* metoda-obserwator musi posiadać parametr opatrzony adnotacją `@Observes` o typie takim jak klasa opisująca zdarzenie,
* obiekt opisujący zdarzenie jest wstrzykiwany jako parametr do metody obsługi zdarzenia.

---

### Zdarzenia – przykład - zdarzenie

Klasa opisująca zdarzenie:

```java
@Getter
@Setter
@AllArgsConstructor
@NoArgsConstructor
public class BookReservation {

    private Book book;

    private User user;

}
```

---

### Zdarzenia – przykład

Wywołanie zdarzenia CDI:

```java
public class BookManager {

    @Inject
    private Event<BookReservation> reservationEvent;

    public String reserveBook() {
        //some standard actions...
        reservationEvent.fire(new BookReservation(book, user));
    }

}
```

Obserwator zdarzenia:

```java
public class ReservationListener {

    public void processReservation(@Observes BookReservation reservation) {
        //react to event
    }

}
```

---

### Model zdarzeń – kwalifikatory

Możemy wykorzystać kwalifikatory, aby generować/reagować na określonego rodzaju zdarzenia.

Obserwator zdarzenia z kwalifikatorami:

```java
public class ReservationListener {

    public void processReservation(@Observes @PayBackUser BookReservation reservation) {
        //react to event
    }

}
```

---

### Model zdarzeń – kwalifikatory - wywołanie

Wywołanie zdarzenia z kwalifikatorami - deklaratywnie:

```java
public class BookManager {

    @Inject
    @PayBackUser
    private Event<BookReservation> reservationEvent;

    public String reserveBook() {
        //some standard actions...
        reservationEvent.fire(new BookReservation(book, user));
    }

}
```

Wywołanie zdarzenia z kwalifikatorami - programistycznie

```java
public class BookManager {

    @Inject
    private Event<BookReservation> reservationEvent;

    public String reserveBook() {
        //some standard actions...
        reservationEvent.select(new AnnotationLiteral<PayBackUser>() {})
                .fire(new BookReservation(book, user));
    }

}
```

---

### Przecięcia - Interceptors

Przecięcia:

* nawiązanie do programowania aspektowego,
* aspekty przecinają aplikację w wielu miejscach,
* np. aspekt kontroli dostępu - wszystkie operacje wymagają sprawdzenia, czy użytkownik ma uprawnienia,
* aspekt transakcyjności - wiele operacji rozpoczyna się od utworzenia transakcji i kończy jej potwierdzeniem lub '
  wycofaniem,
* wykorzystanie przecięć pozwala na rozluźnienie powiązań między komponentami.

---

### Motywacja przecięć - przykład

Bez przecięć kontrola dostępu może wyglądać tak:

```jshelllanguage
public void deleteBook() {
    if (userAllowed(user, Operations.DELETE)) {
        //...
    }
}

public void editBook() {
    if (userAllowed(user, Operations.EDIT)) {
        //...
    }
}

public void reserveBook() {
    if (userAllowed(user, Operations.RESERVE)) {
        //...
    }
}
```

---

### Motywacja przecięć

Rozluźnienie powiązań:

* w czasie wykonania programu logika biznesowa przeplata się z logiką metody `userAllowed()`,
* metoda `userAllowed()` przecina aplikację w wielu miejscach i tego właśnie oczekujemy,
* w programowaniu aspektowym chcemy rozluźnić powiązania między metodą `userAllowed()` i logiką biznesową na poziomie
  kodu źródłowego aplikacji.

---

### Przecięcia – przykład

Standardowo na platformie Java EE przecież można używać za pomocą adnotacji `@Interceptors` nakładanej na klasy lub
metody.

```java
public class BookManager {

    @Interceptors(BookInterceptor.class)
    public String reserveBook() {
        //some standard actions...
    }

}
```

W CDI możliwe jest utworzenie adnotacji stanowiącej wiązanie pomiędzy przecięciem i miejscem, w którym ma być ono wywoływane.

```java
@InterceptorBinding
@Target({ElementType.TYPE, ElementType.METHOD})
@Retention(RetentionPolicy.RUNTIME)
public @interface BookBinding {

}
```

---

### Przecięcia – przykład - imlementacja

Implementacja przecięcia w postaci metody oznaczonej adnotacją `@AroundInvoke` i przyjmującej obiekt `InvocationContext`.

```java
@Interceptor
@BookBinding
public class BookInterceptor implements Serializable {

    @AroundInvoke
    public Object invoke(InvocationContext context) throws Exception {
        //before invoke
        Object result = context.proceed();
        //after invoke
        return result;
    }

}
```

---

### Przecięcia – przykład - wiązanie

Wiązanie można przypisać do poszczególnych metod lub całej klasy - wszystkie metody klasy zostaną wtedy związane z przecięciem.

```java
public class BookManager {

    @BookBinding
    public String reserveBook() {
        //some standard actions...
    }

}
```

---

### Przecięcia - konfiguracja

Konfiguracja wywołania przecięć CDI:

* konieczne jest zadeklarowanie przecięć w pliku `beans.xml`,
* lub oznaczenie ich adnotacją `@Priority()`,
* przecięcia o niższym priorytecie są wywoływane w pierwszej kolejności.

Deklaracja za pomocą `@Priority`:

```java
@Interceptor
@BookBinding
@Priority(1000)
public class BookInterceptor implements Serializable {

}
```

Deklaracja przecięć w `beans.xml`:

```xml
<beans>
  <interceptors>
    <class>interceptors.BookInterceptor</class>
  </interceptors>
</beans>
```

---

### Interceptor JTA

Zarządzanie transakcjami w CDI:

* interceptor `@Transactional`:
* dostępny od Java EE 7,
* interceptor zaimplementowany w ramach JTA;
* poziomy obsługi transakcji:
  * `Transactional.TxType.REQUIRED`,
  * `Transactional.TxType.REQUIRES_NEW`,
  * `Transactional.TxType.MANDATORY`,
  * `Transactional.TxType.SUPPORTS`,
  * `Transactional.TxType.NOT_SUPPORTED`,
  * `Transactional.TxType.NEVER`.

---

### Dekoratory

Wzorzec projektowy Dekorator zaimplementowany tak, aby rozluźnić powiązania między:

* obiektem dekorowanym,
* dekoratorem,
* i kodem, który z nich korzysta.

---

### Dekoratory – przykład

Dekorator:

```java
@Decorator
public class UserNotifierSignatureDecorator implements UserNotifier, Serializable {

    @Inject
    @Delegate
    @Any
    private UserNotifier notifier;

    @Override
    public void notifyUser(String msg) {
        String decoratedMsg = msg + "--\n Send automatically from MegaLibrary";
        notifier.notifyUser(decoratedMsg);
    }

}
```

---

### Dekoratory – konfiguracja

Użycie dekoratorów CDI:

* `@Delegate` - określa punkt wstrzyknięcia dekorowanego obiektu (ang. delegate injection point),
* dekorator musi posiadać dokładnie jeden punkt wstrzyknięcia dekorowanego obiektu,
* nie jest konieczna żadna zmiana w miejscu, gdzie używamy obiektów, które mają być dekorowane,
* dekorowane obiekty są wstrzykiwane tak jak dotychczas,
* kontener CDI automatycznie udekoruje wszystkie obiekty pasujące do kwalifikatorów użytych w punkcie wstrzyknięcia
  dekorowanego obiektu.

---

### Dekoratory - priorytety

Użycie dekoratorów CDI:

* może istnieć wiele dekoratorów dla tego samego interfejsu,
* konieczne jest określenie, w jakiej kolejności dekoratory będą wywoływane,
* inna kolejność może skutkować zupełnie innym zachowaniem,
* nie da się aktywować dekoratora CDI bez równoczesnego określenia jego priorytetu, nie istnieje żadna konstrukcja
  składniowa, która by na to pozwalała.

---

### Dekoratory - konfiguracja

Dwa sposoby aktywacji dekoratora:

* przez deskryptor `beans.xml`,
* przez adnotację `@Priority` na klasie dekoratora,
* dekoratory o niższym priorytecie są wywoływane w pierwszej kolejności.

Deklaracja za pomocą `@Priority`:

```java
@Decorator
@Priority(1000)
public class UserNotifierSignatureDecorator implements UserNotifier, Serializable {

}
```

Deklaracja dekoratora w `beans.xml`:

```xml
<beans>
  <decorators>
    <class>decorators.UserNotifierSignatureDecorator</class>
  </decorators>
</beans>
```

---

### Stereotypy

Stereotypy można traktować jako sposób na połączenie zbioru cech w jedną adnotację dla łatwiejszego użycia.

Definicja stereotypu:

```java
@Named
@RequestScoped
@Stereotype
@Target({TYPE, METHOD, FIELD})
@Retention(RUNTIME)
public @interface Model {

}
```

Wykorzystanie stereotypu:

```java
@Model
public class Bean {
    //...
}
```

---

### CDI w aplikacji Java SE

Wykorzystanie CDI w aplikacjach Java SE wymaga implementacji, która na to pozwala. Przykładem jest `weld-se`:

* dodanie zależności,
* wymagany `beans.xml`,
* możliwość uruchomienia przez dostarczoną metodę `main`:
  * `org.jboss.weld.environment.se.StartMain.main(String[] args)`,
  * nasłuchiwanie na zdanie `ContainerInitialized` jako punkt wejścia;
* możliwość ręcznego uruchomienia kontenera.

```xml
<dependency>
    <groupId>org.jboss.weld.se</groupId>
    <artifactId>weld-se-core</artifactId>
    <version>5.1.2.Final</version>
</dependency>
```

---

### CDI w Java SE - zakres funkcjonalności

W aplikacjach Java SE możemy korzystać z:

* wstrzykiwania (`@Inject`) i kwalifikatorów,
* zdarzeń,
* dekoratorów,
* przecięć,
* callbacków `@PostConstruct` i `@PreDestroy`,
* stereotypów,
* zasięgów:
  * `@ApplicationScoped`,
  * `@Dependent`,
  * `@Singleton`,
  * `@ThreadScoped`.

---

### CDI w Java SE - zasięg wątku

Zasięg wątku:

* pozwala na ograniczenie zasięgu beana do jednego wątku,
* wymaga użycia dekoratora obiektów `Runnable`:

```xml
<beans>
  <decorators>
    <class>org.jboss.weld.environment.se.threading.RunnableDecorator</class>
  </decorators>
</beans>
```

---

### Pomocne źródła

Specyfikacja:

* <https://jakarta.ee/specifications/cdi/4.0/jakarta-cdi-spec-4.0>

Implementacja:

* <https://docs.jboss.org/weld/reference/latest/en-US/html/part5.html>

Uzupełnienie informacji:

* <https://eclipse-ee4j.github.io/jakartaee-tutorial/#introduction-to-jakarta-contexts-and-dependency-injection>
* <https://www.baeldung.com/cdi-event-notification>
* <https://www.baeldung.com/jee-cdi-vs-ejb-singleton>
* <https://www.baeldung.com/cdi-interceptor-vs-spring-aspectj>
