# Narzędzia i Aplikacje Jakarta EE

## Jakarta Persistence - Zagadnienia Zaawansowane

Michał Wójcik

---

### Dziedziczenie w klasach encyjnych

Opis problemu:

* model domenowy może wykorzystywać hierarchię dziedziczenia:
  * szczególnie dla złożonych domen biznesowych, np. w aplikacjach korporacyjnych;
* model domenowy jest utrwalany w bazie danych,
* konieczna jest obsługa relacji dziedziczenia w czasie odwzorowania obiektowo-relacyjnego.

---

### Dziedziczenie w JPA

JPA obsługuje dwa scenariusze użycia:

* klasa bazowa nie jest samodzielną encją
  * w klasie bazowej: `@MappedSuperclass`,
  * w klasach pochodnych: brak dodatkowej konfiguracji,
  * pola klasy bazowej odzwierciedlane jako dodatkowe kolumny w tabeli klasy pochodnej;
* klasa bazowa jest samodzielną encją:
  * w klasie bazowej: `@Inheritance(strategy = ...)` dla określenia rodzaju odwzorowania,
  * `@DiscriminatorColumn(name = "type")` dla określenia nazwy kolumny określającej typ klasy w modelu domenowym,
  * w klasie bazowej i klasach pochodnych: `@DiscriminatorValue("book")` dla określenia wartości dla kolumny `type`.

---

### Dziedziczenie w JPA - klasa bazowa

Klasa bazowa nie jest samodzielną encją:

```java
@Getter
@Setter
@MappedSuperclass
public abstract class PaperCollection {

    @Id
    @GeneratedValue
    private Integer id;

    private String title;

}
```

```java
@Getter
@Setter
@Entity
public class Book extends PaperCollection {

    private String author;

    private String cover;

}
```

```java
@Getter
@Setter
@Entity
public class Comic extends PaperCollection {

    private String author;

    private String ilustrator;

}
```

---

### Dziedziczenie w JPA - samodzielna encja

Klasa bazowa jest samodzielną encją:

```java
@Getter
@Setter
@Entity
@Inheritance(strategy = InheritanceType.JOINED)
@DiscriminatorColumn(name = "type")
@DiscriminatorValue("book")
public class Book {

    @Id
    @GeneratedValue
    private Integer id;

    private String title;

    private String author;

}
```

```java
@Getter
@Setter
@Entity
@DiscriminatorValue("comic")
public class Comic extends Book {

    private String ilustrator;

}
```

---

### Strategia dziedziczenia

Dla dziedziczenia z wykorzystaniem encji jako klasy bazowej dostępne są 3 strategie odwzorowania:

* `SINGLE_TABLE`:
  * pojedyncza tabela dla całej hierarchii dziedziczenia;
* `TABLE_PER_CLASS`:
  * odrębna tabela dla każdej klasy konkretnej w hierarchii dziedziczenia,
  * każda z tabel zawiera kolumny dla pól dziedziczonych oraz pól samej klasy;
* `JOINED`:
  * wspólna tabela z kolumnami dla pól klasy bazowej,
  * dodatkowe tabele dla pól klas pochodnych.

---

### Wersjonowanie obiektów encyjnych

Problem edycji tego samego obiektu encyjnego przez różnych użytkowników:

* systemy z wielodostępem, np. aplikacje korporacyjne,
* czas od pobrania encji z bazy danych do zapisu zmian może być długi:
  * czas spędzony przez użytkownika w formularzu edycji,
  * kolejny użytkownik może rozpocząć edycję encji, zanim poprzedni zakończy swoją;
* bez wersjonowania: nadpisywanie danych:
  * najnowszy UPDATE wygrywa.

---

### Blokowanie optymistyczne

JPA oferuje podstawowy model wersjonowania encji z optymistycznym blokowaniem (ang. optimistic locking)

* optymistycznie zakładamy, że konflikt nie wystąpi:
  * dwóch użytkowników nie rozpocznie edycji równolegle;
* jeśli jednak wystąpi, zablokujemy możliwość nadpisania danych o wyższym numerze wersji,
* przydatne, gdy konflikty występują rzadko:
  * większość zapisów kończy się powodzeniem,
  * sporadyczne odrzucenia nie są uciążliwe dla użytkowników,
  * aplikacja powinna pomóc użytkownikowi w scaleniu danych z konfliktujących wersji.

---

### Blokowanie pesymistyczne

JPA oferuje również pesymistyczny model blokowania:

* pesymistycznie zakładamy, że nastąpi konflikt,
* aby mu zapobiec, blokujemy możliwość rozpoczęcia edycji, jeśli inny użytkownik ją rozpoczął,
* informujemy użytkownika o zaistniałej sytuacji.

---

### Blokowanie optymistyczne - wykorzystanie

Blokowanie optymistyczne:

```java
@Getter
@Setter
@Entity
public class Book {

    @Id
    @GeneratedValue
    private int id;

    @Version
    private int version;

}
```

---

### Blokowanie pesymistyczne - przykład

```java
private class BookService {

    @PersistenceContext
    private EntityManager em;

    @Transactional
    public void lock() {
        Book book = em.createQuery("...")
                .setLockMode(LockModeType.PESSIMISTIC_WRITE)
                .setMaxResults(1)
                .getSingleResult();
    }

}
```

```java
private class BookService {

    @PersistenceContext
    private EntityManager em;

    @Transactional
    public void lock() {
        Book book = em.find(Book.class, 1, LockModeType.PESSIMISTIC_READ);
    }

}
```

---

### Blokowanie pesymistyczne - uwagi

Uwagi:

* `PESSIMISTIC_READ` może być eskalowany do `PESSIMISTIC_WRITE`,
* mechanizm działa w ramach transakcji, która nie powinna być sztucznie wydłużana,
* blokowanie dla celów edycji w widoku powinno być realizowane niestandardowymi mechanizmami na poziomie warstwy biznesowej.

---

### Blokowanie pesymistyczne - blokady

| `LockModeType`                            | `PESSIMISTIC_READ`                   | `PESSIMISTIC_WRITE`                                         |
 |-------------------------------------------|--------------------------------------|-------------------------------------------------------------|
| typ                                       | SHARED LOCK                          | EXCLUSIVE LOCK                                              |
| odczyt bez uzyskiwania dodatkowej blokady | tak                                  | nie                                                         |
| dirty reads                               | nie                                  | nie                                                         |
| non-repeatable reads                      | nie                                  | nie                                                         |
| w celu aktualizacji                       | uzyskać `PESSIMISTIC_WRITE`          | aktualizacja możliwa                                        |
| uzyskanie blokady                         | nikt nie posiada `PESSIMISTIC_WRITE` | nikt nie posiada `PESSIMISTIC_WRITE` lub `PESSIMISTIC_READ` |

---

### Read phenomena

Zjawiska dotyczące odczytu danych w transakcji, które mogłyby być zmienione przez inną transakcję:

* dirty reads,
* non-repeatable reads,
* phantom reads.

---

### Dirty reads

**Dirty read** - niezatwierdzona zależność (uncommitted dependency):

```sql
INSERT INTO BOOKS (id, title) VALUES (1, 'Ruda Sfora');
```

```sql
SELECT title FROM books where id = 1;
-- UPDATE books SET title = 'Ruda Sfora 2'
--     where id = 1;
SELECT title FROM books where id = 1;
-- UPDATE books SET title = 'Ruda Sfora 3'
--     where id = 1;
```

```sql
-- SELECT title FROM books where id = 1;
UPDATE books SET title = 'Ruda Sfora 2'
    where id = 1;
-- SELECT title FROM books where id = 1;
UPDATE books SET title = 'Ruda Sfora 3'
    where id = 1;
```

---

### Dirty reads cd

**Dirty read** - niezatwierdzona zależność (uncommitted dependency):

```sql
INSERT INTO BOOKS (id, title) VALUES (1, 'Ruda Sfora');
```

```sql
SELECT title FROM books where id = 1;
-- UPDATE books SET title = 'Ruda Sfora 2'
--     where id = 1;
SELECT title FROM books where id = 1;
-- ROLLBACK;
```

```sql
-- SELECT title FROM books where id = 1;
UPDATE books SET title = 'Ruda Sfora 2'
    where id = 1;
-- SELECT title FROM books where id = 1;
ROLLBACK;
```

---

### Non-repeatable reads

**Non-repeatable reads**:

```sql
INSERT INTO BOOKS (id, title) VALUES (1, 'Ruda Sfora');
```

```sql
SELECT title FROM books where id = 1;
-- UPDATE books SET title = 'Ruda Sfora 2'
--     where id = 1;
-- COMMIT;
SELECT title FROM books where id = 1;
COMMIT;
```

```sql
-- SELECT title FROM books where id = 1;
UPDATE books SET title = 'Ruda Sfora 2'
    where id = 1;
COMMIT;
-- SELECT title FROM books where id = 1;
-- COMMIT;
```

---

### Phantom reads

**Phantom reads**:

```sql
INSERT INTO BOOKS (id, title) VALUES (1, 'Ruda Sfora');
```

```sql
SELECT * FROM books;
-- INSERT INTO BOOKS (id, title)
--     VALUES (2, 'Ruda Sfora 2');
-- COMMIT;
SELECT title FROM books;
COMMIT;
```

```sql
-- SELECT * FROM books;
INSERT INTO BOOKS (id, title)
    VALUES (2, 'Ruda Sfora 2');
COMMIT;
-- SELECT title FROM books;
-- COMMIT;
```

---

### Poziomy izolacji transakcji

Poziomy izolacji transakcji zależą od możliwości wystąpienia zjawisk.

| poziom izolacji    | dirty reads | non-repeatable reads | phantom reads |
|--------------------|-------------|----------------------|---------------|
| `READ UNCOMMITTED` | X           | X                    | X             |
| `READ COMMITTED`   | -           | X                    | X             |
| `REPEATABLE READ`  | -           | -                    | X             |
| `SERIALIZABLE`     | -           | -                    | -             |

---

### Poziomy izolacji transakcji - konfiguracja

JTA (Java Transaction API) pozwala na zdefiniowanie propagowania transakcji, ale nie pozwala na zdefiniowanie poziomu izolacji.

Większość serwerów SQL domyślnie korzysta z `READ_COMMITTED` (MySQL `REPEATABLE_READ`).

```xml
<property name="openjpa.jdbc.TransactionIsolation" value="repeatable-read"/>
```

```xml
<property name="hibernate.connection.isolation" value="REPEATABLE_READ"/>
```

---

### Propagowanie transakcji

Konfiguracja CDI:

```jshelllanguage
@Transactional(Transactional.TxType.MANDATORY)
public void save(Invoice entity) {

}
```

Konfiguracja EJB:

```jshelllanguage
@TransactionAttribute(TransactionAttributeType.MANDATORY)
public void save(Invoice entity) {
    
}
```

Propagowanie transakcji:

* `MANDATORY` - transakcja musi już istnieć,
* `NEVER` - transakcja nie może istnieć,
* `NOT_SUPPORTED` - transakcja, jeśli istnieje, jest zawieszona,
* `REQUIRED` - transakcja, jeśli nie istnieje, jest tworzona,
* `REQUIRES_NEW` - transakcja, jeśli istnieje, jest zawieszona, tworzy zawsze nową,
* `SUPPORTS` - transakcja może istnieć.

---

### Automatyczna modyfikacja encji

JPA pozwala na automatyczne wykonywanie metod z poziomi encji odpowiadających zdarzeniom z cyklu życia:

* `@PrePersist`,
* `@PreRemove`,
* `@PreUpdate`.

```java
@Getter
@Setter
@Entity
public class Book {

    @Id
    @GeneratedValue
    private int id;

    private LocalDate creation;

    @PrePersist
    private void update() {
        creationDate = LocalDate.now();
    }

}
```

---

### Zapytania w JPQL

**JPQL**:

* składnia nawiązuje do języka SQL
  * łatwy do nauki,
  * zwięzły;
* zapytania w postaci ciągów znaków (`String`):
  * parametry czasu wykonania;
* zapytania `@NamedQuery` weryfikowane w czasie wdrażania aplikacji na serwer,
* podatny na literówki,
* brak podpowiadania składni (stwierdzenie prawie aktualne),
* uciążliwy, gdy zapytanie jest budowane dynamicznie.

---

### Zapytania w Criteria API

**Critera API**: API umożliwiające budowanie zapytań bazodanowych z silnym typowaniem:

* bez używania Stringów,
* wchodzi w skład specyfikacji Javy EE
* kontrola na etapie kompilacji projektu
* podpowiadanie składni w czasie budowania zapytań:
  * poszczególne elementy zapytania dodawane poprzez wywołanie metod obiektów typu builder;
* wykorzystuje statyczny metamodel opisujący strukturę klas encyjnych.

---

### Metamodel

```java
@Entity
public class Book {

    @Id
    private Integer id;

    private String title;

    @Enumerated(STRING)
    private Cover cover;

    @Temporal(TIMESTAMP)
    private Date publishDate;

    @ManyToMany
    private List<Author> authors;

    @ManyToOne
    private User owner;
}
```

```java
@StaticMetamodel(Book.class)
public abstract class Book_ {

    public static volatile
    SingularAttribute<Book, Cover> cover;

    public static volatile
    SingularAttribute<Book, User> owner;

    public static volatile
    SingularAttribute<Book, Date> publishDate;

    public static volatile
    SingularAttribute<Book, Integer> id;

    public static volatile
    SingularAttribute<Book, String> title;

    public static volatile
    ListAttribute<Book, Author> authors;
}
```

---

### Główne klasy Criteria API

Do budowania zapytań wykorzystuje się kilka interfejsów zdefiniowanych w Criteria API:

* `CriteriaBuilder`,
* `CriteriaQuery<T>`,
* `Root<T>`,
* `Join`,
* `ListJoin`,
* `Subquery`,
* `Path`,
* `Predicate`,
* `Expression`.

---

### CriteriaBuilder

**CriteriaBuilder** - umożliwia budowanie:

* zapytań (`CriteriaQuery`),
* predykatów (interfejs `Predicate`),
  * np. predykaty do wykorzystania w sekcji WHERE zapytania;
* kryteriów do sortowania (interfejs `Order`):
  * do wykorzystania w sekcji `ORDER BY` zapytania;
* wyrażeń (interfejs `Expression`)
  * m.in.: odwołania do pól encji, wywołania funkcji agregujących (np. `count()`), funkcji narzędziowych (np. `currentDate()`).

```jshelllanguage
CriteriaBuilder builder = em.getCriteriaBuilder();
```

---

### CriteriaQuery

**CriteriaQuery**- umożliwia skonfigurowanie głównych sekcji zapytania:

* `select()`/`multiselect()` – określenie wartości zwracanych przez zapytania (odpowiednik `SELECT`),
* `where()` – warunki dopasowania zwracanych rekordów (odpowiednik `WHERE`),
* `groupBy()` – grupowanie rekordów (odpowiednik `GROUP BY`),
* `having()` - – odpowiednik `HAVING`,
* `orderBy()` – kryteria sortowania wyników (odpowiednik `ORDER BY`),
* `distinct()` – odpowiednik `DISTINCT`.

---

### Root

**Root**:

* reprezentuje korzeń drzewa encji, które są trawersowane w zapytaniu
* umożliwia wykonywanie złączeń:
  * Metody `join()` – odpowiednik `JOIN` w języku JPQL
  * Metody `fetch()` – odpowiednik `JOIN FETCH`;
* trawersując związki między encjami w złączeniach rozbudowujemy drzewo encji.

---

### Przykładowe zapytanie

```jshelllanguage
public List<Book> findAllBooks() {
    //SELECT b FROM Book b
    CriteriaBuilder builder = em.getCriteriaBuilder();
    CriteriaQuery<Book> query = builder.createQuery(Book.class);
    Root<Book> book=query.from(Book.class);
    query.select(book);
    return em.createQuery(query).getResultList();
}
```

---

### Criteria API

Wykorzystanie Criteria API:

* w prostych zapytaniach wysiłek związany z użyciem Criteria API jest niewspółmierny do zysków:
  * JPQL jest bardziej czytelny i bardziej zwięzły w przypadku zapytań o statycznej strukturze, w których zmieniają
    się tylko wartości parametrów;
* w aplikacjach biznesowych często zachodzi potrzeba dynamicznego budowania zapytań
  * np.: filtrowanie po wielu polach równocześnie z możliwością definiowania warunków logicznych przez użytkownika.

---

### Metamodel - generowanie

Wykorzystanie metamodelu:

* na szczęście metamodelu nie trzeba budować ręcznie,
* do modułu zawierającego klasy encyjne należy dodać odpowiednią zależność,
* klasy metamodelu zostaną wygenerowane w czasie budowania aplikacji.

```xml
<dependency>
    <groupId>org.hibernate</groupId>
    <artifactId>hibernate-jpamodelgen</artifactId>
    <scope>provided</scope>
</dependency>
```

---

### Logowanie zapytań SQL

Konfiguracja po stronie aplikacji (`persistence.xml`):

* Hibernate:

```xml
<properties>
    <property name="hibernate.show_sql" value="true"/>
    <property name="hibernate.format_sql" value="true"/>
</properties>
```

* EclipseLink:

```xml
<properties>
    <property name="eclipselink.logging.level.sql" value="FINE"/>
    <property name="eclipselink.logging.parameters" value="true"/>
</properties>
```

* OpenJPA:

```xml
<properties>
    <property name="openjpa.Log" value="DefaultLevel=WARN,Runtime=INFO,Tool=INFO,SQL=TRACE"/>
</properties>
```

---

### Logowanie zapytań SQL - WildFly

Zaleca się konfigurację poziomów logowania na serwerze zamiast użycia pliku `persistence.xml`.

Hibernate na serwerze Wildfly (`standalone/configuration/standalone.xml`):

* logowanie zapytań SQL:

```xml
<subsystem xmlns="urn:jboss:domain:logging:7.0">
    <logger category="org.hibernate.SQL">
        <level name="DEBUG"/>
    </logger>
</subsystem>
```

* logowanie wartości parametrów zapytań SQL:

```xml
<logger category="org.hibernate.type.descriptor.sql">
    <level name="TRACE"/>
</logger>
```

* logowanie informacji o grupowaniu zapytań:

```xml
<logger category="org.hibernate.engine.jdbc.batch">
    <level name="DEBUG"/>
</logger>
```

---

### WildFly - ustawienia logowania

Globalne ustawienie logów:

```xml
<subsystem xmlns="urn:jboss:domain:logging:7.0">
    <console-handler name="CONSOLE">
        <level name="TRACE"/>
        <formatter>
            <named-formatter name="COLOR-PATTERN"/>
        </formatter>
    </console-handler>
</subsystem>
```

---

### Logowanie zapytań SQL - OpenLiberty

EclipseLink na serwerze OpenLiberty (`server.xml`):

```xml
<logging traceSpecification="eclipselink.sql=all"
         traceFileName="stdout"
         traceFormat="BASIC"/>
```

---

### Grupowanie zapytań

**Grupowanie zapytań** (batch queries):

* popularne implementacje JPA przy ustawieniach domyślnych nie grupują zapytań SQL:
  * każdy `INSERT`/`UPDATE` wykonywany jako osobne zapytanie,
  * narzut po stronie bazy danych na parsowanie zapytań;
* w przypadku wielu zapisów w jednej transakcji grupowanie znacząco skraca czas obsługi,
* konfigurację grupowania może określić w pliku `persistence.xml`.

---

### Grupowanie zapytań (Hibernate)

Grupowanie zapytań:

* maksymalna liczba zgrupowanych zapytań:

```xml
<properties>
    <property name="hibernate.jdbc.batch_size" value="50"/>
</properties>
```

* szeregowanie operacji (na podstawie wartości klucza głównego):

```xml
<properties>
    <property name="hibernate.order_inserts" value="true"/>
    <property name="hibernate.order_updates" value="true"/>
</properties>
```

* obsługa grupowania dla wersjonowanych encji:

```xml
<properties>
    <property name="hibernate.jdbc.batch_versioned_data" value="true"/>
</properties>
```

---

### Grupowanie zapytań (EclipseLink)

Grupowanie zapytań:

* włączenie grupowania zapytań:

```xml
<property name="eclipselink.jdbc.batch-writing" value="JDBC"/>
```

* maksymalna liczba zgrupowanych zapytań:

```xml
<property name="eclipselink.jdbc.batch-writing.size" value="500"/>
```

---

### Grupowanie zapytań - uwagi

Uwagi do grupowania zapytań:

* grupowanie zapytań nie działa dla:

```jshelllanguage
@GeneratedValue(strategy = GenerationType.IDENTITY)
@ID
private int id;
```

* grupowanie zapytań powinno działać dla:

```jshelllanguage
@GeneratedValue(strategy = GenerationType.SEQUENCE)
@ID
private int id;
```

---

### Grupowanie zapytań select (Hibernate)

Grupowanie zapytań:

* liczba obiektów wczytywanych jednorazowo w czasie ładowania powiązanych encji:

```xml
<property name="hibernate.default_batch_fetch_size" value="20"/>
```

* Liczba wierszy odczytywanych jednorazowo z kursora bazodanowego dla zapytania `SELECT`:

```xml
<property name="hibernate.jdbc.fetch_size" value="50"/>
```

---

### Grupowanie zapytań select (EclipseLink)

Grupowanie zapytań:

* liczba obiektów wczytywanych jednorazowo w czasie ładowania powiązanych encji:

```jshelllanguage
return em.createQuery("select b from Book b",Book.class)
        .setHint("eclipselink.batch","b.author")
        .getResultList();
```

* Liczba wierszy odczytywanych jednorazowo z kursora bazodanowego dla zapytania `SELECT`:

```jshelllanguage
return em.createQuery("select b from Book b",Book.class)
        .setHint("eclipselink.batch","b.author")
        .setHint("eclipselink.BATCH_SIZE","100")
        .getResultList();
```

---

### Problem 1+N zapytań

**Problem 1+N zapytań** - pobranie jednym zapytaniem `N` rekordów, a następnie za pomocą `N` zapytań pobranie rekordów
będących w związku.

---

### Grafy encji

**Grafy encji** - (entity graphs) sposób ładowania powiązanych obiektów encyjnych można również
określić za pomocą grafu encji:

* domyślny graf encji wynika z wartości atrybutów `fetch` w adnotacjach `@OneToOne`, `@OneToMany` itd.:
  * nie zaleca się używania strategii `FetchType.EAGER`,
  * jeśli użyto wyłącznie strategii `FetchType.LAZY`, to domyślny graf encji nie zawiera żadnych związków;
* pożądany graf encji należy podać jako wskazówkę (hint) dla JPA w czasie budowania zapytania,
* związki występujące w grafie są ładowane w ramach jednego zapytania,
* pozostałe związki są ładowane w zależności od użytej wskazówki:
  * `jakarta.persistence.loadgraph` – pozostałe związki zgodnie z grafem domyślnym,
  * `jakarta.persistence.fetchgraph` – pozostałe związki nie są ładowane;
* UWAGA: biblioteka Hibernate traktuje opcję `fetchgraph` tak samo jak `loadgraph`:
  * <https://hibernate.atlassian.net/browse/HHH-8776>.

---

### Definicja i wykorzystanie grafu

Definicja grafu:

```jshelllanguage
@NamedEntityGraphs({
        @NamedEntityGraph(
                name = Invoice.Graphs.WITH_ADDRESS,
                attributeNodes = {@NamedAttributeNode("address")}
        )
})
public class Invoice {
    
}
```

Wykorzystanie grafu:

* dla obiektu klasy `Query`:

```jshelllanguage
TypedQuery<Invoice> query = em.createQuery("SELECT a FROM Invoice a", Invoice.class);
EntityGraph graph = em.getEntityGraph(Invoice.Graphs.WITH_ADDRESS);
query.setHint("jakarta.persistence.loadgraph",graph);
List<Invoice> invoices = query.getResultList();
```

* w metodzie `EntityManager.find()`:

```jshelllanguage
EntityGraph entityGraph = em.getEntityGraph(Invoice.Graphs.WITH_ADDRESS);
Map<String, Object> map = Map.of("jakarta.persistence.loadgraph", entityGraph);
Invoice invoice = em.find(Invoice.class, id, map);
```

---

### Pomocne źródła

Specyfikacja:

* <https://jakarta.ee/specifications/persistence/3.1/>

Dokumentacja:

* <https://hibernate.org/orm/documentation/5.6/>.
* <https://hibernate.org/orm/documentation/6.3//>.

---

### Pomocne źródła - uzupełnienie

Uzupełnienie informacji:

* <https://eclipse-ee4j.github.io/jakartaee-tutorial/#persistence>
* <https://www.baeldung.com/hibernate-inheritance>
* <https://www.baeldung.com/hibernate-tips-how-to-map-an-inheritance-hierarchy-to-one-table>
* <https://www.baeldung.com/jpa-optimistic-locking>
* <https://www.baeldung.com/jpa-pessimistic-locking>
* <https://www.baeldung.com/hibernate-criteria-queries>
* <https://www.baeldung.com/jpql-hql-criteria-query>
* <https://www.baeldung.com/jpa-and-or-criteria-predicates>
* <https://www.baeldung.com/hibernate-criteria-queries-metamodel>
* <https://www.baeldung.com/jpa-entity-graph>
* <https://www.baeldung.com/hibernate-fetchmode>
* <https://www.baeldung.com/database-auditing-jpa>
