# Narzędzia i Aplikacje Jakarta EE

## Bean Validation

Michał Wójcik

---

### Mechanizm Bean Validation

**Bean Validation**:

* uniwersalny mechanizm walidacji danych we wszystkich warstwach aplikacji:
  * warstwa interfejsu, np. JSF, JAX-RS,
  * warstwa biznesowa, np. EJB, CDI,
  * warstwa dostępu do danych, np. JPA;
* jednorazowo definiowane walidatory i reguły walidacji do wykorzystania we wszystkich warstwach:
  * nie ma potrzeby duplikowania logiki walidującej;
* pozwala na automatyczną walidację zawartości dowolnych beanów,
* pozwala na tworzenie własnych klasy walidujących, dostępnych później przez adnotacje,
* pozwala walidować pojedyncze właściwości lub całe obiekty.

---

### Specyfikacja Bean Validation

**Bean Validation**:

* specyfikacja,
* Hibernate Validator - implementacja referencyjna.

---

### Standardowe walidatory

| Adnotacja             | Typ pola               | Opis                                                                                       |
|-----------------------|------------------------|--------------------------------------------------------------------------------------------|
| `@NotNull`            | Dowolny                | Wartość inna niż null                                                                      |
| `@Null`               | Dowolny                | Wartość równa null                                                                         |
| `@Size(min=X, max=Y)` | Ciągi znaków, kolekcje | Rozmiar w przedziale [X, Y]                                                                |
| `@Pattern(regex=X)`   | Ciągi znaków           | Ciąg znaków pasuje do wyrażenia regularnego                                                |
| `@Min(value=X)`       | Liczby                 | Wartość liczbowa minimum X                                                                 |
| `@Max(value=X)`       | Liczby                 | Wartość liczbowa maksimum X                                                                |
| `@Future`             | Daty                   | Data w przyszłości                                                                         |
| `@Past`               | Daty                   | Data w przeszłości                                                                         |
| `@AssertTrue`         | Boolean                | Flaga ma wartość true                                                                      |
| `@AssertFalse`        | Boolean                | Flaga ma wartość false                                                                     |
| `@NotEmpty`           | Ciągi znaków, kolekcje | Element nie zawiera wartości null i nie jest pusty                                         |
| `@NotBlank`           | Ciągi znaków           | Element nie zawiera wartości null i nie jest pusty, ignoruje białe znaki na końcu wartości |
| `@Email`              | Ciągi znaków           | Ciąg znaków zawiera poprawny adres e-mail                                                  |

---

### Przykład

```java
@Entity
public class Book {
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    private Integer id;

    @NotBlank
    private String title;

    @Past
    @Temporal(TemporalType.TIMESTAMP)
    private Date publishDate;

    @Size(min = 1)
    @ManyToMany
    private List<Author> authors = new ArrayList<>();

}
```

---

### Własny walidator

Definicja własnego walidatora obejmuje dwa elementy:

* Klasę z interfejsem `ConstraintValidator`:
  * zawiera logikę weryfikującą poprawność przekazywanej wartości;
* adnotację znakującą:
  * pozwala wskazać pola mające podlegać walidacji,
  * opatrzona adnotacją `@Constraint`,
  * może posiadać dodatkowe parametry do sprecyzowania reguł walidacji.

---

### Adnotacja znakująca

```java
@Target({ElementType.FIELD, ElementType.METHOD})
@Retention(RetentionPolicy.RUNTIME)
@Constraint(validatedBy = InPastValidator.class)
@Documented
public @interface InPast {

    String message() default "date should be in past";

    Class<?>[] groups() default {};

    Class<? extends Payload>[] payload() default {};

    String date() default "";

}
```

---

### Meta-adnotacje

Definicja adnotacji:

* semantykę adnotacji definiuje się za pomocą innych adnotacji,
* `@Target` - określa jakie elementy mogą być oznaczane adnotacją,
* `@Retention` - określa, jak długo będzie dostępna informacja o obecności adnotacji:
  * `RUNTIME` - w czasie działania aplikacji,
  * `CLASS` - w skompilowanych klasach,
  * `SOURCE` - tylko w kodzie źródłowym;
* `@Documented` - wystąpienia adnotacji zostaną udokumentowane w klasach, które ich używają.

---

### Meta-adnotacja @Constraint

Meta-adnotacja `@Constraint`:

* adnotacje opatrzone meta-adnotacją `@Constraint` są rozpoznawane jako adnotacje znakujące walidatorów:
  * atrybut `validatedBy` określa klasę z logiką walidatora;
* adnotacja znakująca musi zawierać pola:
  * `message` - komunikat informujący o błędzie walidacji:
    * może odnosić się do klucza wielojęzycznego zasobu (`ValidationMessages.properties` w domyślnym pakiecie);
  * `groups` - lista grup, w których reguła walidacji jest aktywna,
  * `payload` – dodatkowa informacja do wykorzystania przez dewelopera (nie jest używana przez Bean Validation API).

---

### Klasa walidująca

```java
public class InPastValidator implements ConstraintValidator<InPast, Date> {

    public static final String DATE_PATTERN = "yyyy-MM-dd";

    private Date date = new Date();

    @Override
    public void initialize(InPast constraintAnnotation) {
        if (!constraintAnnotation.date().equals("")) {
            try {
                date = new SimpleDateFormat(DATE_PATTERN)
                        .parse(constraintAnnotation.date());
            } catch (ParseException ex) {
            }
        }
    }

    @Override
    public boolean isValid(Date value, ConstraintValidatorContext context) {
        return date.after(value);
    }

}
```

---

### Przykłady użycia

```java
@Entity
public class Book {
    
    //...

    @InPast(date = "2013-10-19")
    @Temporal(TemporalType.TIMESTAMP)
    private Date publishDate;

}
```

---

### Walidacja powiązanych pól

Walidacja więcej niż jednego pola na raz:

* w wielu sytuacjach obiekty posiadają pola o powiązanych ze sobą wartościach, np.:
  * nowe hasło oraz powtórz nowe hasło (w formularzu rejestracji/zmiany hasła),
  * miasto oraz kod pocztowy (w formularzu zamówienia);
* walidacja wymaga dostępu do wartości wszystkich powiązanych pól,
* Bean Validation umożliwia zadeklarowanie walidatora na poziomie klasy:
  * z dostępem do wszystkich składowych obiektów tej klasy.

---

### Przykład: formularz rejestracji

```java
public class RegistrationForm {

    @NotBlank
    String login;

    @NotBlank
    String password;

    @NotBlank
    String passwordRepeat;

}
```

---

### Adnotacja znakująca i klasa walidujące

```java
@Target({TYPE})
@Retention(RUNTIME)
@Constraint(validatedBy = RepeatPasswordValidator.class)
@Documented
public @interface RepeatPassword {

    String message() default "Hasła muszą być takie same";

    Class<?>[] groups() default {};

    Class<? extends Payload>[] payload() default {};

}
```

```java
public class RepeatPasswordValidator
        implements ConstraintValidator<RepeatPassword, RegistrationForm> {

    @Override

    public void initialize(RepeatPassword constraintAnnotation) {

    }

    @Override
    public boolean isValid(RegistrationForm value, ConstraintValidatorContext context) {
        return value.getPassword().equals(value.getPasswordRepeat());
    }

}
```

---

### Wykorzystanie walidatora

```java
@RepeatPassword
public class RegistrationForm {

    @NotBlank
    String login;

    @NotBlank
    String password;

    @NotBlank
    String passwordRepeat;

}
```

---

### Grupowanie reguł walidacji

**Grupowanie reguł walidacji**:

* ten sam obiekt może wymagać innych reguł walidacji na różnych etapach przetwarzania:
  * dane wpisywane w formularzu przez użytkownika,
  * pola o wartościach wyliczanych w warstwie logiki biznesowej,
  * weryfikacja wszystkich pól przed zapisem do bazy danych;
* Bean Validation pozwala na przypisywanie reguł walidacji do grup,
* w czasie walidacji należy wskazać, która grupa ma zostać zweryfikowana,
* grupy deklarowane są jako interfejsy znakujące (interfejsy bez metod):
  * z braku bardziej adekwatnego mechanizmu w języku.

Przykład:

* grupa dla informacji wprowadzanych przez użytkownika w formularzu na stronie:

```java
public interface UserInput {
}
```

* grupa dla informacji dodawanych przez wydawcę:

```java
public interface PublisherAddons {
}
```

---

### Przypisywanie grup

```java
@Entity
public class Book {

    @Id
    private Integer id;

    @NotBlank
    //@NotBlank(groups = Default.class) //default group
    private String title;

    @Past(groups = PublisherAddons.class)
    @Temporal(TemporalType.TIMESTAMP)
    private Date publishDate;

    @Pattern(regexp = "...", groups = PublisherAddons.class)
    private String isbn;

    @Size(min = 1, groups = UserInput.class)
    @ManyToMany
    private List<Author> authors = new ArrayList<>();

}
```

---

### Walidacja wybranych grup

```jshelllanguage
ValidatorFactory factory = Validation.buildDefaultValidatorFactory();
Validator validator = factory.getValidator();
```

Tylko domyślna grupa:

```jshelllanguage
Set<ConstraintViolation<Book>> errors = validator.validate(book);
```

Grupa domyślna oraz `UserInput`:

```jshelllanguage
validator.validate(book, Default.class, UserInput.class);
```

Tylko grupa `PublisherAddons`:

```jshelllanguage
validator.validate(book, PublisherAddons.class);
```

---

### Bean Validation a JSF

W JSF adnotacje Bean Validation nałożone na własności (lub ich pola) podłączone pod kontrolki są przetwarzane
automatycznie.

Wybór grup walidacyjnych:

```xml
<f:validateBean validationGroups="pl.edu.pg.eti.UserInput"/>
```

Walidacja całego obiektu:

```xml
<f:validateWholeBean value="#{book}" validationGroups="pl.edu.pg.eti.UserInput"/>
```

Uwaga:

* podczas walidacji całego obiektu domyślna grupa jest ignorowana, należy wykorzystać jakaś własną,
* pola obiektu podpięte pod interfejs muszą deklarować tę samą grupę,
* funkcjonalność musi być wprost włączona w `web.xml`.

```xml
<context-param>
    <param-name>jakarta.faces.validator.ENABLE_VALIDATE_WHOLE_BEAN</param-name>
    <param-value>true</param-value>
</context-param>
```

---

### Bean Validation a JAX-RS

W JAX-RS adnotacje Bean Validation mogą być nałożone bezpośrednio na parametry (query, form i path):

```jshelllanguage
@POST
@Path("/books")
public void createBook(@NotBlank @FormParam("title") String title) {
    //...
}
```

Adnotacje nałożone na reprezentację ciała żądania (na poziomie klasy i pól) są przetwarzane przy wykorzystaniu adnotacji
`@Valid`:

```jshelllanguage
@POST
@Path("/books")
public void createBook(@Valid Book book) {
    //...
}
```

---

### Wykorzystanie poza Jakarta EE

```xml
<dependency>
    <groupId>org.hibernate.validator</groupId>
    <artifactId>hibernate-validator</artifactId>
    <version>8.0.0.Final</version>
</dependency>
```

---

### Pomocne źródła

Specyfikacja:

* <https://jakarta.ee/specifications/bean-validation/3.0/jakarta-bean-validation-spec-3.0>

Dokumentacja:

* <https://docs.jboss.org/hibernate/stable/validator/reference/en-US/html_single/>.

Uzupełnienie informacji:

* <https://eclipse-ee4j.github.io/jakartaee-tutorial/#bean-validation>
* <https://www.baeldung.com/java-validation>
* <https://www.baeldung.com/java-bean-validation-constraint-composition>
* <https://www.baeldung.com/java-bean-validation-not-null-empty-blank>
* <https://www.baeldung.com/java-validation-list-annotations>
* <https://www.baeldung.com/jersey-bean-validation>
* <https://www.baeldung.com/javax-validation-method-constraints>
* <https://www.baeldung.com/bean-validation-container-elements>
* <https://www.baeldung.com/spring-boot-bean-validation>
* <https://www.baeldung.com/rest-localized-validation-messages>
