# Narzędzia i Aplikacje Jakarta EE

## Jakarta RESTful WS - Zagadnienia Zaawansowane

Michał Wójcik

---

### Dodatkowe informacje kontekstowe

Za pomocą adnotacji `@Context` można wstrzyknąć obiekty klas:

* `HttpServletRequest` - obiekt reprezentujący żądanie,
* `HttpServletResponse` - obiekt reprezentujący odpowiedź,
* `UriInfo` - informacje o adresie i parametrach żądania,
* `HttpHeaders` - nagłówki żądania,
* `ServletContext` - kontekst servletu obsługującego żądanie,
* `SecurityContext` - informacje o uprawnieniach klienta:
  * `getUserPrincipal()`,
  * `isUserInRole()`.

---

### HATEOAS w JAX-RS

Obecna specyfikacja **JAX-RS** nie dostarcza wygodnego rozwiązania *out of the box* dla **HATEOAS**:

* istnieją pomocne rozwiązania w implementacjach JAX-RS:
  * nie są częścią standardu,
  * wymagają zależności do konkretnej implementacji standardu,
  * niekoniecznie tej, która jest na naszym serwerze aplikacji;
* istnieją zewnętrzne rozwiązania:
  * dodatkowe zależności,
  * zewnętrzne biblioteki, które niekoniecznie będą rozwijane w przyszłości.

---

### Budowanie odnośników - przykład

Odnośniki można budować ręcznie, przykładowo:

```jshelllanguage
@POST
@Path("/books")
public Response addBook(Book book) {
    //...save book...
  return Response.created(URI.create("/books/" + book.getId())).build();
}
```

Zwrócony adres (np.: `/books/17`) wskazuje na metodę z adnotacją `@Path("books/{id})`:

```jshelllanguage
@GET
@Path("books/{id}")
public Response getBookById(@PathParam("id") Integer id) {
    Book book = service.findBook(id);
    if (book != null) {
        return Response.ok(book).build();
    } else {
        return Response.status(Response.Status.NOT_FOUND).build();
    }
}
```

---

### Budowanie odnośników

**Ręczne budowanie** odnośników:

* kłopotliwe,
* sklejanie ciągów znaków z fragmentami adresu:
  * `"/authors/" + author.getId() + "/books/" + book.getId()`;
* podatne na literówki, błędy dewelopera,
* trudne w utrzymaniu:
  * jeśli zmienimy routing (adnotacje `@Path`), to musimy poprawić w kodzie wszystkie miejsca, gdzie adresy były
    budowane ręcznie.

---

### Odwrócony routing

Przy budowaniu odnośników przydatny jest tzw. **odwrócony routing** (ang. reverse routing):

* routing - odwzorowanie typu: adres URI → akcja kontrolera,
* reverse routing - odwzorowanie w drugą stronę: akcja kontrolera → adres URI,
* należy wskazać klasę kontrolera oraz jego metodę, aby uzyskać adres URI, który powoduje jej wywołanie,
* mechanizm *reverse routing* jest obsługiwany przez klasę narzędziową `UriBuilder`.

---

### UriBuilder

**Budowanie odnośnika**:

* pobranie adresu obsługiwanego przez klasę:
  * bieżący adres: `/books/`,
  * adnotacja `@Path` na poziomie klasy `BookResource`;
* pobranie adresu obsługiwanego przez metodę:
  * bieżący adres: `/books/{id}`,
  * łączy zawartość adnotacji `@Path` na poziomie klasy `BookResource` oraz na poziomie metody`getBookById()`.

```jshelllanguage
UriBuilder.fromResource(BookResource.class);
```

```jshelllanguage
UriBuilder.fromResource(BookResource.class)
        .path(BooksResource.class, "getBookById")
```

---

### UriBuilder - parametry

**Budowanie odnośnika**:

* jeśli adres zawiera parametry, należy podać ich wartości w wywołaniu metody `build()`:
  * wynikowy adres `books/17`.

```jshelllanguage
UriBuilder.fromResource(BookResource.class)
        .path(BooksResource.class, "getBookById")
        .build(book.getId());
```

---

### HATEOAS - przykład - zasób główny

```json
{
  "links": [
    {
      "rel": "self",
      "href": "/api"
    },
    {
      "rel": "/api/relations/books",
      "href": "/api/books"
    },
    {
      "rel": "/api/relations/authors",
      "href": "/api/authors"
    }
  ],
  "name": "Library"
}
```

---

### HATEOAS - przykład - kolekcja

```json
{
  "links": [
    {
      "rel": "self",
      "href": "/api/books"
    }
  ],
  "books": [
    {
      "title": "Pan Lodowego Ogrodu",
      "links": [
        {
          "rel": "self",
          "href": "/api/professions/Pan_Lodowego_Ogrodu"
        }
      ]
    }
  ]
}
```

---

### HATEOAS - przykład

Przykład głównego zasobu aplikacji:

* `links` wskazuje na odnośniki do innych zasobów,
* struktura inspirowana standardem ATOM,
* każdy odnośnik opisany przez `href` i `rel`,
* `href` opisuje adres zasobu,
* `rel` opisuje relację,
* `self` jest relacją zdefiniowaną w rejestrze relacji IANA,
* pozostałe relacje przedstawione w postaci odnośnika do zasobu je opisującego.

---

### HAL

**Hypermedia Application Language** (HAL) - otwarta specyfikacja opisującą strukturę zasobów typu REST z wykorzystaniem
relacji hipertekstowych.

Minimum to wprowadzenie relacji `self` wskazującej na aktualnie przeglądany zasób:

```json
{
  "id": 2,
  "title": "Pan Lodowego Ogrodu tom 1",
  "publishDate": "2011-02-05",
  "cover": "HARD",
  "_links": {
    "self": {
      "href": "http://example.com/library/api/books/2"
    }
  }
}
```

Typ MIME odpowiedzi: `application/json+hal` lub `application/xml+hal`.

---

### HAL - Zawieranie innych zasobów

Inne zasoby zawarte w zasobie umieszcza się jako `_embedded`:

```json
{
  "id": 2,
  "title": "Pan Lodowego Ogrodu tom 1",
  "publishDate": "2011-02-05",
  "cover": "HARD",
  "_links": {
    "self": {
      "href": "http://example.com/library/api/books/2"
    }
  },
  "_embedded": {
    "authors": [
      {
        "id": 2,
        "name": "Jarosław",
        "surname": "Grzędowicz",
        "_links": {
          "self": {
            "href": "http://example.com/library/api/authors/2"
          }
        }
      }
    ]
  }
}
```

---

### HAL - kolekcje

Kolekcje są reprezentowane w ramach `_embedded`:

```json
{
  "_embedded": {
    "books": [
      {
        "id": 1,
        "title": "Ruda Sfora",
        "publishDate": "2013-10-11",
        "cover": "SOFT",
        "_links": {
          "self": {
            "href": "http://example.com/library/api/books/1"
          }
        }
      }
    ]
  },
  "_links": {
    "self": {
      "href": "http://example.com/library/api/books"
    }
  }
}
```

---

### HAL - relacje do innych zasobów

Zasób może zawierać relacje do innych zasobów zamiast (lub dodatkowo) zawierać je w sobie:

```json
{
  "id": 2,
  "title": "Pan Lodowego Ogrodu tom 1",
  "publishDate": "2011-02-05",
  "cover": "HARD",
  "_links": {
    "self": {
      "href": "http://example.com/library/api/books/2"
    },
    "authors": {
      "href": "http://example.com/library/api/books/2/authors"
    }
  }
}
```

---

### HAL - definiowanie operacji

Zasób może definiować inne operacje niż tylko pobranie (`GET`):

```json
{
  "id": 2,
  "title": "Pan Lodowego Ogrodu tom 1",
  "publishDate": "2011-02-05",
  "cover": "HARD",
  "_links": {
    "self": {
      "href": "http://example.com/library/api/books/2"
    },
    "authors": {
      "href": "http://example.com/library/api/books/2/authors"
    },
    "addReview": {
      "href": "http://example.com/library/api/books/2/reviews",
      "method": "POST"
    }
  }
}
```

---

### HAL - paginacja

W przypadku kolekcji (zwłaszcza dużych) warto zadbać o paginację:

```json
{
  "_embedded": {
    "books": [
    ]
  },
  "_links": {
    "self": {
      "href": "http://example.com/library/api/books?page=5"
    },
    "first": {
      "href": "http://example.com/library/api/books?page=0"
    },
    "last": {
      "href": "http://example.com/library/api/books?page=10"
    },
    "next": {
      "href": "http://example.com/library/api/books?page=6"
    },
    "previous": {
      "href": "http://example.com/library/api/books?page=4"
    }
  }
}
```

---

### HAL - paginacja i właściwości

W ramach paginacji mogą być załączone dodatkowe własności:

```json
{
  "_embedded": {
    "books": [
    ]
  },
  "count": 4,
  "total": 43,
  "_links": {
    "self": {
      "href": "http://example.com/library/api/books?page=5"
    },
    "first": {
      "href": "http://example.com/library/api/books?page=0"
    },
    "last": {
      "href": "http://example.com/library/api/books?page=10"
    },
    "next": {
      "href": "http://example.com/library/api/books?page=6"
    },
    "previous": {
      "href": "http://example.com/library/api/books?page=4"
    }
  }
}
```

---

### HAL - opis relacji

```json
{
  "_links": {
    "self": {
      "href": "http://example.com/library/api/books/2"
    },
    "curies": [
      {
        "href": "/api/relations/{rel}",
        "name": "library",
        "templated": true
      }
    ]
  },
  "_embedded": {
    "library:authors": [
      {
        "Grzędowicz",
        "_links": {
          "self": {
            "href": "http://example.com/library/api/authors/2"
          }
        }
      }
    ]
  }
}
```

---

### HAL - opis relacji - uwagi

Opis relacji w HAL:

* relacje niezdefiniowane w katalogu relacji IANA powinny być odpowiednio opisane,
* odnośniki do zasobu opisującego relację,
* zalecana składania CURIE (Compact URI),
* każda relacja posiada prefiks,
* prefiksy zdefiniowane w kluczu `curies`.

---

### JAX-RS - obsługa wyjątków

**Obsługa wyjątków**:

* należy unikać *wyciekania* wyjątków biznesowych do: klienta RESTowego API
* nieprzechwycone wyjątki powodują wysłanie odpowiedzi typu 500: Internal server error,
* deweloper aplikacji klienckiej wchodzi na naszego bug trackera i wypełnia zgłoszenie błędu;
* wyjątki biznesowe należy przetłumaczyć na odpowiednie kody odpowiedzi HTTP, np.:
  * `NoResultException` → 404: Not found
  * `ValidationException` → 422: Unprocessable Entity
  * `SecurityException`/`EJBAccessException` → 403: Forbidden.

---

### Odwzorowanie wyjątków - przykład

**Odwzorowanie**:

* odwzorowanie wyjątków biznesowych na kody HTTP można zdefiniować za pomocą obiektów odwzorowujących,
* odwzorowanie ma postać klasy implementującej interfejs `ExceptionMapper<T>`.

```java
@Provider
public class EJBAccessExceptionMapper implements ExceptionMapper<EJBAccessException> {

    @Override
    public Response toResponse(EJBAccessException exception) {
        return Response.status(FORBIDDEN).build();
    }
    
}
```

---

### Odwzorowanie wyjątków

**Odwzorowanie**:

* wyjątki, dla których istnieją odwzorowanie, nie muszą być ręcznie przechwytywane i obsługiwane w kontrolerach,
* jeśli taki wyjątek zostanie wyrzucony z metody kontrolera, to JAX-RS automatycznie przetłumaczy go na kod odpowiedzi
  HTTP,
* zastosowanie odwzorowań upraszcza konstrukcję kontrolerów,
* i chroni przez "wyciekaniem" wyjątków biznesowych do klienta.

---

### Wyjątki w JAX-RS

**Obsługa wyjątków**:

* JAX-RS w szczególny sposób obsługuje wyjątki dziedziczące po `WebApplicationException`,
* wyjątki tego typu zawierają informację o kodzie odpowiedzi HTTP do zwrócenia,
* JAX-RS odsyła do klienta kod odpowiedzi odczytany z wyjątku.

```jshelllanguage
throw new WebApplicationException(Response.Status.NOT_FOUND);
```

---

### Wyjątki WebApplicationException

Istnieje hierarchia wyjątków dziedziczących po `WebApplicationException` z predefiniowanymi kodami:

* `NotFoundException` (odpowiedź 404),
* `ForbiddenException` (odpowiedź 403),
* `NotAuthorizedException` (odpowiedź 401),
* `BadRequestException` (odpowiedź 400),
* `NotAllowedException` (405: Method Not Allowed),
* `NotSupportedException` (415: Unsupported Media Type),
* `NotAcceptableException` (odpowiedź 406),
* `RedirectionException` (odpowiedź 3xx).

---

### Wyjątki WebApplicationException - przykład

Przykład - obsługa odpowiedzi 404: Not found:

```jshelllanguage
@GET
@Path("/{id}")
public Book getBookById(@PathParam("id") Integer id) {
    Book book = bookService.findBook(id);
    if (book == null) {
        throw new NotFoundException();
    }
    return book;
}
```

Dla porównania - z wykorzystaniem klasy Response:

```jshelllanguage
@GET
@Path("/{id}")
public Response getBookById(@PathParam("id") Integer id) {
    Book book = bookService.findBook(id);
    return book == null 
            ? Response.status(NOT_FOUND).build() 
            : Response.ok(book).build();
}
```

---

### Własne wyjątki

**Obsługa wyjątków**:

* można również definiować własne wyjątki dziedziczące po `WebApplicationException`,
* Istnieje pokusa, aby we własnych wyjątkach biznesowych dziedziczyć po `WebApplicationException`:
  * nie będzie potrzeby definiowania dla nich odwzorowań,
  * wyjątki biznesowe będą przez to posiadały zależności od JAX-RS,
  * złamanie podziału na warstwy - moduł webowy może być zależny od warstwy biznesowej, ale warstwa biznesowa nie
    powinna być zależna od modułu webowego,
  * dziedziczenie jest możliwe tylko we własnych klasach,
  * wyjątki kontenera EJB/bibliotek i tak wymagają obsługi.

---

### Pomocne źródła

Specyfikacja:

* <https://datatracker.ietf.org/doc/html/draft-kelly-json-hal-11>

Uzupełnienie informacji:

* <https://www.baeldung.com/spring-hateoas-tutorial>
* <https://www.baeldung.com/spring-rest-hal>

Publikacje:

* Michał Wójcik, *Implementacja wykrywalnych usług typu REST na platformie Jakarta EE*,
  *Uczenie maszynowe i systemy rozproszone*,
  <https://pbc.gda.pl/dlibra/publication/109835/edition/98484>.
