# Narzędzia i Aplikacje Jakarta EE

## Enterprise Java Beans - Zagadnienia Zaawansowane

Michał Wójcik

---

### Zadania automatyczne

**Zadania automatyczne**:

* wykonanie pojedynczego zadania:
  * po określonym czasie,
  * jednorazowa opóźniona akcja;
* wykonanie cyklicznego zadania:
  * cyklicznie o określonej porze,
  * cykliczna akcja,
  * np.: sprzątanie zasobów.

---

### Timer Service - upływ czasu

Timer programistyczny:

```java
@Singleton
@Startup
public class DelayedInit {

    @Resource
    private TimerService timerSerivce;

    @PostConstruct
    private void init() {
        timerService.createTimer(1000, "ping");
    }

    @Timeout
    public void execute(Timer timer) {
        System.out.println(timer.getInfo());
    }

}
```

---

### Timer Service - harmonogram

Automatyczna akcja wykonywana co minutę:

```java
@Singleton
@Startup
public class Scheduled {

    @Schedule(minute = "*/1", hour = "*", persistent = false)
    public void execute() {
        //...
    }

}
```

---

### Wywołania asynchroniczne

Asynchronicznie wywołania metod:

* zamiast blokować żądanie, od razu zwracana jest wartość ("obietnica" zrobienia zadania),
* wykonywane w osobnym wątku.

Tworzenie wywołań asynchronicznych:

* oznaczenie metody za pomocą `@Asynchronous`,
* metody asynchronicznie muszą zwracać `void` lub implementację `Future<T>`,
* metoda niezwłocznie po wywołaniu zwraca wynik,
* właściwa praca odbywa się w tle,
* sprawdzanie, czy wywołanie zostało anulowane, poprzez `SessionContext#wasCancelled()`.

---

### Wywołania asynchroniczne - serwis

Przykład metody asynchronicznej:

```java
@Stateless
public class HelloService {

    @Resource
    private SessionContext ctx;

    @Asynchronous
    public Future<String> longHello() {
        String result = null;
        while (isRunning()) {
            if (ctx.wasCancelled()) {
                // clean up
                break;
            }
            // do sth awesome
            if (isDone()) {
                result = "hello";
                break;
            }
        }
        return new AsyncResult<>(result);
    }

}
```

---

### Wywołania asynchroniczne - widok

Przykład widoku pozwalającego użytkownikowi sterować długotrwałym zadaniem:

```java
@ViewScoped
@Named
public class HelloView {

    @EJB
    private HelloService service;

    private Future<String> future;

    public void start() {
        future = service.longHello();
    }

    public boolean isDone() {
        return future.isDone();
    }

    public void cancel() {
        future.cancel(true);
    }

    public String getResult() {
        return future.get();
    }
}
```

---

### Java Message Service

Java Message Service:

* komunikacja "luźno" powiązana,
* komunikacja asynchroniczna,
* dostarczenie wiadomości jest gwarantowane,
* dostępna jest opcja niegwarantowanego dostarczania wiadomości,
* może być wykorzystywana nie tylko z Java EE,
* administrator tworzy:
  * cel (destination):
  * fabryki połączeń (connection factories).

---

### Java Message Service cd

Java Message Service:

* nadawca i odbiorca nie muszą o sobie wiedzieć,
* elementy:
  * JMS provider - system wymiany wiadomości, implementacja Java EE musi go zawierać,
  * JMS clients - konsumenci i producenci wiadomości,
  * messages - obiekty wymieniane w wiadomościach,
  * administered objects - obiekty wykorzystywane przez klientów:
    * cele,
    * fabryki połączeń.

---

### JMS - cele

Cele:

* Kolejka:
  * model p2p,
  * tylko jeden konsument na wiadomość,
  * brak zależności czasowych,
  * potwierdzenie odebrania wymagane. przetworzenie wiadomości.
* Temat:
  * model publish/subscribe
  * możliwość istnienia wielu konsumentów na wiadomość,
  * zależności czasowe,
  * potwierdzenie niewymagane.

---

### JMS - wymiana wiadomości

Konsumpcja wiadomości:

* synchroniczna: blocking receive,
* asynchroniczna: message listener.

---

### JMS - elementy wiadomości

Wiadomości składają się z:

* nagłówka:
  * część pól ustawiana jest automatycznie,
  * część można ustawić ręcznie;
* właściwości:
  * mogą służyć do filtrowania wiadomości,
  * np. typy, kategorie, wiadomości;
* ciała wiadomości:
  * zależy od typu wiadomości.

---

### JMS - typy wiadomości

Wiadomości JMS:

* dostępnych jest sześć typów wiadomości:
  * `TextMessage` - przesyłanie obiektów String,
  * `MapMessage` - mapy String → dowolny typ prymitywny,
  * `BytesMessage` - wiadomość binarna,
  * `StreamMessage` - strumień wartości prymitywnych typów,
  * `ObjectMessage` - serializowalny obiekt,
  * `Message` - zawiera tylko nagłówek i właściwości (bez ciała wiadomości);
* możliwe jest ustawienie priorytetów wiadomości i czasu ważności,
* JMS ma własny mechanizm transakcji dostępny dla klientów, którzy nie wykonują się w serwerze aplikacji.

---

### Message-Driven EJB

Message-Driven EJB:

* pozwala na asynchroniczne przetwarzanie komunikatów,
* klienty nie korzystają z nich przy pomocy interfejsów,
* ten typ EJB składa się tylko z klasy beana,
* nie posiadają stanu,
* wykorzystują Java Message Service (JMS).

---

### Klasy Message-Driven EJB

Klasy Message-Driven EJB:

* posiadają adnotację `@MessageDriven` i implementują interfejs `MessageListener`,
* implementują jedną metodę `onMessage`, która jest wywoływana po otrzymaniu wiadomości - logika biznesowa,
* może zawierać metody wywoływane w określonych etapach cyklu życia beana.

---

### Message-Drive EJB

Przykład konsumenta wiadomości:

```java
@MessageDriven(activationConfig = {
        @ActivationConfigProperty(
                propertyName = "destination",
                propertyValue = "jms/BookQueue"),
        @ActivationConfigProperty(
                propertyName = "destinationType",
                propertyValue = "javax.jms.Queue")
})
public class ReadMessageMDB implements MessageListener {

    public void onMessage(Message message) {
        TextMessage textMessage = (TextMessage) message;
    }

}
```

---

### Wysyłanie wiadomości

Wysyłanie wiadomości:

```java
@Stateless
public class MessageProducer {

    @Resource(lookup = "jms/BookFactory")
    private ConnectionFactory connectionFactory;

    @Resource(lookup = "jms/BookQueue")
    private Queue queue;

    public void send(String msg) {
        try (Connection connection = connectionFactory.createConnection();
             Session session = connection.createSession();
             MessageProducer producer = session.createProducer(queue)) {

            TextMessage message = session.createTextMessage();
            message.setText("");
            producer.send(message);

        } catch (Throwable ex) {
            //...
        }
    }

}
```

---

### Odbieranie wiadomości

Odbieranie wiadomości:

```java
@Stateless
public class MessageConsumer {

    @Resource(lookup = "jms/BookFactory")
    private ConnectionFactory connectionFactory;

    @Resource(lookup = "jms/BookQueue")
    private Queue queue;

    public void receive() {
        try (Connection connection = connectionFactory.createConnection();
             Session session = connection.createSession();
             MessageConsumer consumer = session.createConsumer(queue)) {

            Message msg = consumer.receive();

        } catch (Throwable ex) {
            //...
        }
    }

}
```

---

### Cykl życia EJB sterowanego komunikatami

![width80](./assets/img/ejb-messages.svg)

---

### Transakcje - konfiguracja

W przypadku transakcji zarządzanych przez kontener to kontener rozpoczyna i kończy transakcje, bazując na odpowiednich
adnotacjach. Dostępne są następujące opcje transakcji:

* `Required` (domyślne ustawienie) - jeśli dostępna to korzysta z transakcji klienta, jak nie to tworzy nową,
* `RequiresNew` - jeśli dostępna do zawiesza transakcję klienta i tworzy własną,
* `Mandatory` - wyrzuca wyjątek, jeśli transakcji klienta jest niedostępna,
* `NotSupported` - jeśli dostępna do zawiesza transakcję klienta,
* `Supports` - wykonuje się zarówno gdy transakcji klienta jest dostępna i niedostępna, nie tworzy własnej,
* `Never` - wyrzuca wyjątek, jeśli transakcji klienta jest dostępna.

---

### Transakcje - ograniczenia

Transakcje zarządzane przez kontener:

* można wyspecyfikować typ transakcji dla:
  * całego beana,
  * wybranych metod biznesowych a w beanach typu message-driven dla metod `onMessage`;
* anulowanie transakcji zarządzanych przez kontener następuje:
  * przez wywołanie `setRollbackOnly()` na `EJBContext` - dostęp do `EJBContext` można uzyskać adnotacją `@Resource`,
  * automatycznie, jeśli wystąpi wyjątek systemowy;
* używając tego typu transakcji nie wolno korzystać z:
  * metod: `commit`, `setAutoCommit`, `rollback` klasy `java.sql.Connection`,
  * metod klasy `javax.transaction.UserTransaction`.

---

### Transakcje - przykład

Transakcje zarządzane przez kontener:

```java
@Stateless
public class ContainerManagedBean {

}
```

Transakcje zarządzanie przez beana:

```java
@Stateless
@TransactionManagement(TransactionManagementType.BEAN)
public class ManagedBean {

}
```

---

### Bezpieczeństwo deklaratywne w EJB

Bezpieczeństwo deklaratywne w EJB:

* moduł EJB - wymagane role definiowane za pomocą adnotacji na klasach/metodach biznesowych:
  * `@RolesAllowed(...)` - dostęp dla użytkowników o podanych rolach,
  * `@PermitAll` - dostęp dla wszystkich użytkowników,
  * `@DenyAll` - niedostępne dla nikogo;
* zasięg:
  * adnotacja na klasie dotyczy wszystkich metod tej klasy,
  * adnotacje metodach nadpisują adnotacje na klasie.
* wymagane jest użycie adnotacji `@DeclareRoles` w połączeniu z `@RolesAllowed`.

---

### Bezpieczeństwo deklaratywne - przykład

Zabezpieczony bean EJB:

```java
@Stateless
@DeclareRoles({"ADMIN", "USER"})
public class BookService {

    @RolesAllowed("USER")
    public List<Book> findAll() {
        ///...
    }

    @RolesAllowed("ADMIN")
    public void deleteAll() {
        ///...
    }

}
```

---

### Bezpieczeństwo deklaratywne - warianty

Dostępne są dwa warianty definiowania wymaganych ról:

* wariant dozwalający (ang. permissive):
  * jeśli nie zostaną określone role wymagane dla operacji, jest ona dostępna dla wszystkich,
  * `jboss-ejb3.xml`: `missing-method-permissions-deny-access` → `false`;
* wariant restrykcyjny (ang. restrictive):
  * jeśli nie zostaną określone role wymagane dla operacji, nie jest ona dostępna dla nikogo,
  * jeśli deweloper przeoczy jakąś metodę, to nie będzie ona dostępna dla potencjalnych napastników,
  * `jboss-ejb3.xml`: `missing-method-permissions-deny-access` → `true`.

---

### Bezpieczeństwo programistyczne w EJB

Bezpieczeństwo programistyczne w EJB:

* obiekt klasy `SessionContext` udostępnia informacje o użytkowniku:
  * `isCallerInRole()` - sprawdzenie roli,
  * `getCallerPrincipal()` - uzyskanie tożsamości;
* wstrzyknięcie za pomocą adnotacji `@Resource`.

---

### Wykorzystanie `SessionContext`

Zabezpieczony bean EJB:

```java
@Stateless
@DeclareRoles({"ADMIN", "USER"})
public class BookService {

    @Resource
    private SessionContext ctx;

    @RolesAllowed({"ADMIN", "USER"})
    public void saveBook(Book book) {
        boolean isAdmin = ctx.isCallerInRole("ADMIN");
        String login = ctx.getCallerPrincipal().getName();
        boolean isOwner = book.getOwner().getLogin().equals(login);

        if (isAdmin || isOwner) {
            //...
        } else {
            throw new SecurityException("Insufficient permissions");
        }
    }

}
```

---

### Tworzenie aplikacji klienckiej Java EE

Aplikacje kliencie:

* klientem EJB może być dowolna aplikacja napisana w Javie (inne języki i środowiska są możliwe),
* w aplikacji klienckiej można korzystać wyłącznie z interfejsów zdalnych EJB,
* dostęp do EJB uzyskuje się przy pomocy adnotacji lub JNDI:
  * adnotacje działają tylko w klasie głównej aplikacji;
* aplikacja kliencka może mieć deskryptor application-client.xml,
* klasa główna aplikacji i zależności od innych bibliotek są podawane w manifeście JAR-a,
* do uruchomienia aplikacji klienckiej potrzebny jest dostęp do:
  * bibliotek klienckich serwera aplikacji,
  * bibliotek z interfejsami zdalnymi EJB oraz klasami (interfejsami) współdzielonymi w aplikacji
    klienckiej i module EJB.

---

### Uruchamianie aplikacji klienckiej Java EE

Aplikacje kliencie:

* aplikacja kliencka uruchamiana jest w specjalnym kontenerze klienckim,
* sposób uruchamiania aplikacji jest zależny od serwera aplikacji:
  * często dołączany jest skrypt lub program uruchamiający;
* jeśli aplikacja kliencka jest uruchamiana na innym komputerze niż serwer aplikacji, to musi mieć otwarty dostęp do
  odpowiednich portów serwera aplikacji, m.in.:
  * usługi JNDI,
  * portów pozwalających na wołanie `EJBContext`,
  * ...
* jest możliwe uruchomienie aplikacji klienckiej bez kontenera klienckiego, ale wymaga to dodatkowych działań
  konfiguracyjnych oraz dołączenia bibliotek klienckich serwera aplikacji.

---

### EJB jako model komponentowy

EJB jako model komponentowy:

* przed wprowadzeniem CDI, w Java EE istniały dwa główne modele komponentów:
  * komponenty JSF (`@ManagedBeans`) - warstwa interfejsu,
  * komponenty EJB - warstwa logiki biznesowej;
* CDI dostarcza pojedynczy, uniwersalny model komponentów co użycia we wszystkich warstwach aplikacji,
* kolejne wydania Javy EE integrują coraz więcej elementów wokół CDI.

---

### EJB vs CDI

CDI oferuje większość najczęściej wykorzystywanych możliwości EJB:

* wstrzykiwanie zależności,
* zarządzanie transakcjami (`@Transactional`),
* obsługę komunikatów (model zdarzeń)
  * od CDI 2.0 również obsługa zdarzeń asynchronicznych (`Event#fireAsync`),
  * alternatywa dla JMS i message-driven EJB do komunikacji wewnątrz aplikacji;
* szereg możliwości niedostępnych wcześniej w komponentach EJB:
  * dekoratory, wiązania przecięć, fabryki, nowe zasięgi.

---

### Ograniczenia CDI

CDI nie pokrywa wszystkich możliwości EJB

* zdalne EJB:
  * w CDI tylko komunikacja wewnątrz aplikacji;
* kolejki komunikatów do integracji z zewnętrznymi systemami:
  * zdarzenia CDI tylko w obrębie aplikacji;
* pule bezstanowych obiektów gotowych do obsługi klientów:
  * mogłyby zostać zrealizowane w CDI za pomocą nowych zasięgów,
  * kwestia kosztu inicjalizacji komponentu, np. JNDI lookup w obrębie klastra serwerów aplikacji.

---

### EJB (zdalne) jako usługi sieciowe

EJB (zdalne) a usługi sieciowe:

* binarny protokół komunikacyjny:
  * wysoka wydajność w porównaniu do protokołów tekstowych,
  * wymaga tej samej technologii po stronie serwera i po stronie klienta,
  * brak interoperacyjności z innymi językami programowania/platformami technologicznymi;
* współcześnie typowo wykorzystujemy usługi w architekturze REST:
  * wysoka interoperacyjność;
* dostępne są też nowoczesne protokoły binarne, zaprojektowane z myślą o interoperacyjności:
  * np. protokół gRPC wykorzystujący format ProtocolBuffers to transferu danych,
  * biblioteki dostępne dla wszystkich popularnych języków.

---

### EJB jako kolejka komunikatów

EJB a kolejki komunikatów:

* wykorzystanie kolejek komunikatów pozwala na rozluźnienie powiązań między komponentami systemu:
  * pożądane szczególnie w dużych aplikacjach;
* broker dostarczy komunikat, nawet jeśli adresat nie był dostępny w chwili jego wysyłania,
* możliwość skalowania systemu przez zwiększenie liczby serwerów obsługujących żądania (ang. worker):
  * co jest przezroczyste z punktu widzenia komponentu emitującego komunikaty,
  * możliwość optymalizacji kosztów w chmurze (np. Spot Instances w chmurze AWS).

---

### EJB jako kolejka komunikatów cd

EJB a kolejki komunikatów:

* nie da się ukryć, że kolejki komunikatów są przydatnym mechanizmem:
  * czy ich implementacja w postaci usługi JMS jest najlepszym wyborem?
* JMS nie jest protokołem komunikacyjnym,
* JMS to API dla aplikacji w języku Java:
  * problem z interoperacyjnością;
* lepiej zastosować uniwersalny protokół komunikacyjny:
  * zrozumiały dla aplikacji zaimplementowanych w różnych językach programowania,
  * RabbitMQ (firma Pivotal),
  * Amazon Simple Queue Service (SQS),
  * ...

---

### EJB dzisiaj

EJB dzisiaj:

* komponenty EJB nie powinny być domyślnym wyborem w nowych projektach:
  * większość możliwości EJB jest dostępna w innych, nowszych rozwiązaniach;
* oczywiście istnieją starsze aplikacje, które wymagają utrzymywania:
  * powstały w czasach, kiedy nie było jeszcze lepszych alternatyw dla EJB,
  * albo architekci zastosowali w nich EJB "z przyzwyczajenia".

---

### Pomocne źródła

Specyfikacja:

* <https://jakarta.ee/specifications/messaging/3.1/jakarta-messaging-spec-3.1>

Uzupełnienie informacji:

* <https://www.baeldung.com/scheduling-in-java-enterprise-edition>
* <https://www.baeldung.com/spring-remoting-jms>
* <https://www.baeldung.com/spring-jms>
* <https://eclipse-ee4j.github.io/jakartaee-tutorial/#messaging>
* <https://eclipse-ee4j.github.io/jakartaee-tutorial/#using-asynchronous-method-invocation-in-session-beans>
* <https://eclipse-ee4j.github.io/jakartaee-tutorial/#using-the-timer-service>
