# Narzędzia i Aplikacje Jakarta EE

## Angular - Reactive Forms

Michał Wójcik

---

### Template-driven forms

**Template-driven forms**:

* wykorzystuje `ngModel`
* import modułu `FormsModule`,
* wykorzystanie `[(ngModel)]` do wiązania danych,
* wykorzystanie `(ngSubmit)` do zatwierdzenia formularza,
* walidacja za pomocą dyrektyw,
* przypisanie kontrolek bezpośrednio do danych,
* dobry do prostych formularzy.

---

### Reactive forms

**Reactive forms**:

* import modułu `ReactiveFormsModule`,
* wiązanie danych za pomocą `formGroup` i `formControl`,
* wykorzystanie `(ngSubmit)` do zatwierdzenia formularza,
* przypisanie kontrolek do obiektów formularza,
* walidacją za pomocą walidatorów przypisanych do obiektów formularza,
* dobry do złożonych i dynamicznych formularzy.

---

### Template-driven vs Reactive

Template-driven vs Reactive Forms:

* Template-driven:
  * łatwe w użyciu,
  * większa zależność od szablonu,
  * mniej kontroli nad formularzem w kodzie,
  * mniejsza elastyczność w testowaniu.
* Reactive:
  * więcej kodu do napisania,
  * lepsza separacja logiki od widoku,
  * łatwiejsze testowanie,
  * lepsza obsługa złożonych formularzy.

---

### Wartość formularza

Jako typ opisujący wartość formularza możemy skorzystać z interfejsu.

```typescript
export interface BookFormValue {
    
  title: string;
  
  isHardcover: boolean;
  
}
```

---

### Opis formularza

W celu opisania silnie typowanego formularza można skorzystać z definicji typu.

```typescript
import { FormControl, FormGroup } from '@angular/forms';

export type BookForm = FormGroup<{
    
  title: FormControl<string>;
  
  isHardcover: FormControl<boolean>;
  
}>;
```

---

### Tworzenie formularza

Aby móc podpiąć formularza pod kontrolki, musi on być wpierw stworzony:

* proste formularze często tworzy się w komponencie,
* do skomplikowanych definiuje się osobne serwisy.

```typescript
export class BookFormService {
    createBookForm(): BookForm {
        return new FormGroup({
            title: new FormControl(''),
            isHardcover: new FormControl(false)
        });
    }
}
```

---

### Obsługa null

Obsługa wartości `null`:

* domyślnie kontrolki stworzonę z początkową wartością będą miały typ `|null`:
  * np.: `new FormControl('ookami')` będzie miał typ `FormControl<string|null>`;
* przy resetowaniu formularza ustawiane są wartości `null`,
* można uniknąć typu `null` oraz resetowanie do wartości domyślnych z użyciem opcji `nonNullable`.

```typescript
export class BookFormService {
    createBookForm(): BookForm {
        return new FormGroup({
            title: new FormControl(
                '',
                {
                    nonNullable: true
                }
            ),
            isHardcover: new FormControl(
                false,
                {
                    nonNullable: true
                }
            )
        });
    }
}
```

---

### Walidatory

Walidatory mogą być wiązane na poziomie kodu bezpośrednio do obiektów formularza.

```typescript
export class BookFormService {
    createBookForm(): BookForm {
        return new FormGroup({
            title: new FormControl(
                '',
                {
                    nonNullable: true,
                    validators: [
                        Validators.required,
                        Validators.pattern('([A-Z].*[.])([ ][A-Z].*[.])*')
                    ]
                }
            ),
            isHardcover: new FormControl(
                false,
                {
                    nonNullable: true
                }
            )
        });
    }
}
```

---

### Domyślne walidatory

Lista domyślnych walidatorów:

* **required**: Sprawdza, czy pole formularza nie jest puste.
* **requiredTrue**: Wymaga, aby wartość pola była `true`. Zazwyczaj używany z polami typu checkbox.
* **minLength**: Sprawdza minimalną długość pola tekstowego.
* **maxLength**: Sprawdza maksymalną długość pola tekstowego.
* **pattern**: Waliduje pole formularza za pomocą wyrażenia regularnego.
* **email**: Sprawdza, czy pole formularza zawiera poprawny adres e-mail.
* **min**: Sprawdza, czy wartość pola jest większa lub równa wartości minimalnej.
* **max**: Sprawdza, czy wartość pola jest mniejsza lub równa wartości maksymalnej.
* **nullValidator**: Zapewnia, że wartość jest różna od null (domyślny walidator).
* **compose**: Tworzy skomponowane walidatory w celu zastosowania wielu walidatorów do pola.
* **composeAsync**: Podobny do `compose`, ale do asynchronicznych walidatorów.

---

### Własne walidatory

Własne walidatory realizuje się poprzez implementację `ValidatorFn`. Najwygodniej wydzielić do tego odpowiednią klasę
z metodami statycznymi (spójne ze standardowym `Validator`).

```typescript
export class CustomValidators {
    
  static exactValue(expected: string): ValidatorFn {
    return (control: AbstractControl): ValidationErrors | null => {
      if (control.value !== expected) {
        return {
          exactValue: {
            value: control.value
          }
        }
      }
      return null;
    }
  }

}
```

---

### Szablon formularza

```angular2html
<form [formGroup]="form" (ngSubmit)="onSubmit()">

    <input id="title" name="title" [formControl]="form.controls.title"/>
  
    <div *ngIf="form.controls.title.invalid && form.controls.title.errors?.['required']">
        Value is required.
    </div>
  
    <div *ngIf="form.controls.title.invalid && form.controls.title.errors?.['pattern']">
        Wrong pattern.
    </div>
  
    <div *ngIf="form.controls.title.invalid && form.controls.title.errors?.['exactValue']">
        Not valid value.
    </div>
  
    <button type="submit">Update</button>
    <button type="button" (click)="form.reset()">Reset</button>
</form>
```

---

### Pomocne źródła

Pomocne źródła:

* <https://angular.io/guide/form-validation>
* <https://angular.io/guide/reactive-forms>
* <https://angular.io/guide/typed-forms>
